"""
labels.py

Einsum-string and ncon Label processing.
"""

from typing import List
import torch

from .generic_functions import first_occurrence_mask


@torch.jit.script
def einsumstr_to_labels(einsum_string : str, device : torch.device=torch.device("cpu")):
    """
    GIVEN : einsum_string (str, numpy.einsum-string)
    GET   : LHS (List[1d-int-torch.tensors] labelling each tensor in einsum_string)
            RHS (1d-int-torch.tensor labelling external indices of output tensor)
            intratraces (List[1d-int-torch.tensors] labelling each tensor, in intra-interal-indices)
    """

    liststring    = einsum_string.replace(" ","").split("->") ## this is a list (at most 2 entries: LHS, RHS)
    if len(liststring)==2: ## RHS given
        LHS = liststring[0].split(",")
        LHS = [torch.tensor([ord(char) for char in word], device=device) for word in LHS]
        RHS =  torch.tensor([ord(char) for char in liststring[1]], device=device)

        global_internal, counts = torch.unique(torch.concat([torch.unique(lhs) for lhs in LHS]), return_counts=True)
        global_internal = global_internal[counts == len(LHS)]
        not_these       = torch.concat([global_internal, RHS]) ## for intraintersect

        intratraces=[] ## over each value of
        for lhs in LHS:
            unique_values, counts = torch.unique(lhs, return_counts=True) ## intra-dupes
            unique_values = unique_values[torch.all(torch.ne(unique_values.unsqueeze(1), not_these.unsqueeze(0)),1)]
            intratraces.append( torch.unique(unique_values) )

    else: #if len(liststring)==1: ## no RHS, go reg. convention (repeats are dummies)
        LHS = liststring[0].split(",") ## should be at most 2-here
        LHS = [torch.tensor([ord(char) for char in word], device=device) for word in LHS]

        ### build RHS-label
        RHS   = liststring[0].replace(",","")
        RHS   = torch.tensor([ord(char) for char in RHS], device=device)
        unique_values, counts = torch.unique(RHS, return_counts=True)
        dupes = unique_values[counts > 1]     # Filter-out duplicate values, gather
        mask  = torch.logical_not(torch.any(torch.eq(RHS.unsqueeze(1), dupes.unsqueeze(0)), 1))
        RHS   = RHS[mask] ## in org. order

        global_internal, counts = torch.unique(torch.concat([torch.unique(lhs) for lhs in LHS]), return_counts=True)
        global_internal = global_internal[counts == len(LHS)]
        not_these       = torch.concat([global_internal, RHS]) ## for intraintersect

        # Perform element-wise XOR comparison
        intratraces=[] ## over each value of
        for lhs in LHS:
            unique_values, counts = torch.unique(lhs, return_counts=True) ## intra-dupes
            dupes = unique_values[counts > 1]
            dupes = dupes[torch.all(torch.ne(dupes.unsqueeze(1), not_these.unsqueeze(0)),1)]
            intratraces.append( torch.unique(dupes) )

    #remove duplicates
    labelA = LHS[0][first_occurrence_mask(LHS[0])]
    #remove intratraces
    labelA = labelA[torch.logical_not(torch.any((labelA.unsqueeze(1) == intratraces[0].unsqueeze(0)), dim=1))]

    #remove duplicates
    labelB = LHS[1][first_occurrence_mask(LHS[1])]
    #remove intratraces
    labelB = labelB[torch.logical_not(torch.any((labelB.unsqueeze(1) == intratraces[1].unsqueeze(0)), dim=1))]
    #remove inter-traces....

    As_frees = torch.any((labelA.unsqueeze(1) == RHS.unsqueeze(0)), dim=1) ## frees
    Bs_frees = torch.any((labelB.unsqueeze(1) == RHS.unsqueeze(0)), dim=1)
    rhs      = torch.concat( (labelA[As_frees], labelB[Bs_frees]) )
    
    onlydumb = torch.logical_and(torch.logical_not(As_frees).unsqueeze(1), torch.logical_not(Bs_frees).unsqueeze(0))
    interdum = (labelA.unsqueeze(1) == labelB.unsqueeze(0)) ## everything 
    adjmatrix= torch.stack( torch.where(( torch.logical_and(onlydumb, interdum)  )) )

    lhs = [labelA, labelB]
    return LHS, RHS, lhs, rhs, intratraces, adjmatrix

##
## NCON LIST
##

@torch.jit.script
def ncon_to_labels(ncon : List[torch.Tensor]):
    """
    GIVEN : einsum_string (np.einsum string)
    GET   : LHS (List[1d-int-torch.tensors] labelling each tensor in einsum_string)
            RHS (1d-int-torch.tensor labelling external indices of output tensor)
            intratraces (List[1d-int-torch.tensors] labelling each tensor, in intra-interal-indices)
    """

    ### build LHS
    LHS = ncon

    ### build RHS-label
    RHS   = torch.concat((LHS))
    unique_values, counts = torch.unique(RHS, return_counts=True)
    dupes = unique_values[torch.logical_or((counts > 1),(unique_values<0))]  # filter-out duplicate values
    mask  = torch.logical_not(torch.any(torch.eq(RHS.unsqueeze(1), dupes.unsqueeze(0)), 1))
    RHS   = RHS[mask] ## in org. order

    global_internal, counts = torch.unique(torch.concat([torch.unique(lhs) for lhs in LHS]), return_counts=True)
    global_internal = global_internal[counts == len(LHS)]
    not_these       = torch.concat([global_internal, RHS]) ## for intraintersect

    intratraces=[]
    for lhs in LHS:
        unique_values, counts = torch.unique(lhs, return_counts=True) ## intra-dupes
        dupes = unique_values[torch.logical_or((counts > 1), (unique_values<0))]
        dupes = dupes[torch.all(torch.ne(dupes.unsqueeze(1), not_these.unsqueeze(0)),1)]
        intratraces.append( torch.unique(dupes) )

    return LHS, RHS, intratraces

## TEST 
#ncon_example = [torch.tensor([1,8,-5,3]), torch.tensor([3,87,3])]

#lhs, rhs, intratr = ncon_to_labels(ncon_example)
#print(lhs)
#print(rhs)
#print(intratr)


"""
dense_intra.py

This contains functions for processing dense tensor intra-operations (slicing & tracing).
"""

import torch
from .generic_functions import first_occurrence_mask, pytorch_delete, iargsort

### given LHS && intra-traces compute reduced/sliced sparse-tensor
@torch.jit.script
def den_post_trans(a, rhs, RHS):
    if torch.numel(RHS)>0:
        rhs=rhs[first_occurrence_mask(rhs)]
        j  = torch.argsort(rhs)
        k  = torch.argsort(RHS)
        ik = iargsort(k)
        return torch.permute(a, [int(i.item()) for i in j[ik]])
    else: ## a is a scalar (cannot permute)
        return a

@torch.jit.script
def den_post_intraTr(a, label):
    """
    GIVEN : a (torch.tensor{dense})
            label (1d-int-torch.tensor)
            intratrace (1d-int-torch.tensor)
    GET   : a (torch.tensor{dense}, intratraced and sliced)
            label (1d-int-torch.tensor, with duplicates and intraTr removed, keeping org. order)
    """
    if torch.numel(label)>0: ## if not empty, its a tensor....
        in_string = ""
        for i in label:
            in_string+=chr(i)

        out_string = ""
        label = label[first_occurrence_mask(label)]
        for i in label:
            out_string+=chr(i)
        return torch.einsum(in_string + "->" + out_string, a) #, label
    else: ## its a scalar return itself
        return a

@torch.jit.script
def den_tensor_intraTr(a, label, intratrace):
    """
    GIVEN : a (torch.tensor{dense})
            label (1d-int-torch.tensor)
            intratrace (1d-int-torch.tensor)
    GET   : a (torch.tensor{dense}, intratraced and sliced)
            label (1d-int-torch.tensor, with duplicates and intraTr removed, keeping org. order)
    """

    in_string = ""
    for i in label:
        in_string+=chr(i)

    out_string = ""
    label = label[first_occurrence_mask(label)]
    for i in label:
        if torch.any(i==intratrace): ## element-wise search
            label = pytorch_delete(label, i==label)
        else:
            out_string+=chr(i)

    return torch.einsum(in_string + "->" + out_string, a)

### TEST
#A    = torch.rand((3,3,3,3,3,3,3,3))
#iqwe = torch.tensor([67, 84, 101, 101, 84, 85, 84, 84])
#print( den_tensor_intraTr(A, iqwe, torch.tensor([])) )


"""
bisum.py

Primary Function of the bisum package, for sparse & dense tensor partial-tracing
"""

import torch

###
from .labels import einsumstr_to_labels, ncon_to_labels
from .tensordot import ss_tensordot , sd_tensordot , sd_tensordot
from .tensordot import ss_tensordot_, sd_tensordot_, ds_tensordot_
from .dense_intra import den_tensor_intraTr, den_post_intraTr, den_post_trans

from .sparse_intra import spa_post_trans, spa_tensor_intraTr, spa_post_intraTr 
from .sparse_intra import spa_post_trans_, spa_tensor_intraTr_, spa_post_intraTr_

def bisum(Rx, a, b):
    """

    """
    if torch.is_tensor(Rx): # and Rx.shape[0]==2:  # is adj.matrix (no post transpose nor slice) 
        if (not a.is_sparse) and (not b.is_sparse): ## both dense
            c = torch.tensordot(a, b, dims=Rx)
        else:
            if (a.is_sparse) and (b.is_sparse):
                c = ss_tensordot(a, b, dims=Rx)
            else:
                c = sd_tensordot(a, b, dims=Rx)

    else:
        if isinstance(Rx, list): # ncon  (no post transpose)  
            LHS, RHS, lhs, rhs, inTr, adjmat = ncon_to_labels(Rx)
        else:
            if isinstance(Rx, str): # einsum ncon_to_labels
                LHS, RHS, lhs, rhs, inTr, adjmat = einsumstr_to_labels(Rx, device=a.device)
            else:
                raise ValueError("tracing instructions are not valid")

        if (not a.is_sparse) and (not b.is_sparse): ## both dense
            a = den_tensor_intraTr(a, LHS[0], inTr[0])
            b = den_tensor_intraTr(b, LHS[1], inTr[1])

            c = torch.tensordot(a, b, dims=adjmat)

            c = den_post_intraTr(c, rhs)
            c = den_post_trans(c, rhs, RHS) #for dense
        else:
            if a.is_sparse and b.is_sparse: ## both sparse
                
                a_index, a_data, a_shape = spa_tensor_intraTr_(a, LHS[0], inTr[0])
                b_index, b_data, b_shape = spa_tensor_intraTr_(b, LHS[1], inTr[1])
                c_index, c_data, c_shape = ss_tensordot_(a_index, a_data, a_shape, b_index, b_data, b_shape, dims=adjmat)
                c_index, c_data, c_shape = spa_post_intraTr_(c_index, c_data, c_shape, rhs)
                c = spa_post_trans_(c_index, c_data, c_shape, rhs, RHS)
                
            else:
                if (a.is_sparse) and (not b.is_sparse):
                    a_index, a_data, a_shape = spa_tensor_intraTr_(a, LHS[0], inTr[0])
                    #a = spa_tensor_intraTr(a, LHS[0], inTr[0]) ##!!!
                    b = den_tensor_intraTr(b, LHS[1], inTr[1])

                    #c = sd_tensordot(a, b, dims=adjmat) ##!!!
                    c = sd_tensordot_(a_index, a_data, a_shape, b, dims=adjmat)

                    c = den_post_intraTr(c, rhs)
                    c = den_post_trans(c, rhs, RHS)

                else: ## a is dense and b is sparse
                    a = den_tensor_intraTr(a, LHS[0], inTr[0])
                    #b = spa_tensor_intraTr(b, LHS[1], inTr[1]) ##!!!
                    b_index, b_data, b_shape = spa_tensor_intraTr_(b, LHS[1], inTr[1])

                    #c = sd_tensordot(a, b, dims=adjmat) ##!!!
                    c = ds_tensordot_(a, b_index, b_data, b_shape, dims=adjmat)
                    
                    c = den_post_intraTr(c, rhs)
                    c = den_post_trans(c, rhs, RHS)
    return c


"""
sparse_intra.py

This contains functions for processing sparse tensor intra-operations (slicing & tracing).
"""

import torch
from .generic_functions import first_occurrence_mask, lexsort, iargsort

### given LHS && intra-traces compute reduced/sliced sparse-tensor

@torch.jit.script
def spa_post_trans(a, rhs, RHS):
    rhs=rhs[first_occurrence_mask(rhs)]
    j = torch.argsort(rhs)
    k = torch.argsort(RHS)
    ik= iargsort(k)
    m = j[ik]

    cc= a._indices()[m,:]
    s = torch.tensor(a.shape)[m]
    n = lexsort( cc )
    return torch.sparse_coo_tensor( cc[:,n] , a._values()[n], [int(i.item()) for i in s])

@torch.jit.script
def spa_post_intraTr(a, label):

    uniques, counts = torch.unique(label, return_counts=True)
    if torch.all(counts==1): ## no duplicates in label
        return a #, label
    else: ## slice and wellorder after then
        index = a._indices()
        data  = a._values()
        shape_= torch.tensor(a.shape)

        ## row-wise slice
        for uni in uniques[counts>1]:
            lookat = torch.where(uni==label)[0]

            expand = index[lookat[0]].unsqueeze(0).expand_as(index[lookat[1:]])
            result = (index[lookat[1:]] == expand)
            I      = torch.all(result, dim=0)
            index  = index[:,I] ## row-wise-chop
            data   = data[I]    ## row-wise-chop
        
        ## column-wise slice, intra-external indices
        keep  = first_occurrence_mask(label)
        index = index[keep] ## chop list-of-tuples
        label = label[keep] ## chop label
        shape_= shape_[keep]
        size  = [int(elem) for elem in shape_]
        
        if torch.numel(index)!=0: ## index is empty
            #print(index.shape)
            i     = lexsort(index)
            index = index[:,i]
            data  = data[i]

            domains = torch.concat([torch.ones_like(label[0].reshape((1)), dtype=torch.int32), torch.any((torch.diff( index , dim=1) != 0), dim=0).long()])

            ## do indices
            index = index[:,domains!=0]

            ## do data
            domains = (torch.cumsum(domains, 0)-1)
            ZZ      = torch.zeros_like(index[0], dtype=a.dtype)
            data    = ZZ.scatter_add(0, domains, data)

            return torch.sparse_coo_tensor(index, data, size) #, label
        else:
            data = torch.unsqueeze(torch.unsqueeze(data, 0),0)
            return torch.sparse_coo_tensor(torch.zeros_like(data[:,:,0], dtype=torch.int64), torch.sum(data), [int(b) for b in torch.ones_like(data[:,0,0])])


@torch.jit.script
def spa_tensor_intraTr(a, label, intratrace):

    uniques, counts = torch.unique(label, return_counts=True)
    if torch.all(counts==1) and torch.numel(intratrace)==0: ## no duplicates in label
        return a #, label
    else: ## slice and wellorder after then
        index = a._indices()
        data  = a._values()
        shape_= torch.tensor(a.shape, device=index.device)

        ## row-wise slice
        for uni in uniques[counts>1]:
            lookat = torch.where(uni==label)[0]

            expand = index[lookat[0]].unsqueeze(0).expand_as(index[lookat[1:]])
            result = (index[lookat[1:]] == expand)
            I      = torch.all(result, dim=0)
            index  = index[:,I] ## row-wise-chop
            data   = data[I]    ## row-wise-chop
        
        ## column-wise slice, intra-external indices
        keep  = first_occurrence_mask(label)
        index = index[keep] ## chop list-of-tuples
        label = label[keep] ## chop label
        shape_= shape_[keep]

        ## column-wise slice, remove intratrace
        if torch.numel(intratrace)>0:
            for j in intratrace: ## each column
                keep   = (label!=j)
                label  = label[keep]
                index  = index[keep]
                shape_ = shape_[keep]
            size = [int(elem) for elem in shape_]
        else:
            size = [int(elem) for elem in shape_]
    
        if index.shape[0]==0: ## no more indices just do full sum of data....
            data = torch.unsqueeze(torch.unsqueeze(data, 0),0)
            return torch.sparse_coo_tensor(torch.zeros_like(data[:,:,0], dtype=torch.int64), torch.sum(data), [int(b) for b in torch.ones_like(data[:,0,0])])
        else: ## do partial-trace/sum
            ### lex-order index array
            i     = lexsort(index)
            index = index[:,i]
            data  = data[i]

            domains = torch.concat([torch.ones_like(label[0].reshape((1)), dtype=torch.int32), torch.any((torch.diff( index , dim=1) != 0), dim=0).long()])

            ## do indices
            index = index[:,domains!=0]

            ## do data
            domains = (torch.cumsum(domains, 0)-1)
            ZZ      = torch.zeros_like(index[0], dtype=a.dtype)
            data    = ZZ.scatter_add(0, domains, data)

            return torch.sparse_coo_tensor(index, data, size) #, label
    
@torch.jit.script
def spa_tensor_intraTr_(a, label, intratrace):
    """ backend-version
    GIVEN   :  ()
    GET     :
    """

    index = a._indices()
    data  = a._values()
    shape_= torch.tensor(a.shape, device=a.device)

    uniques, counts = torch.unique(label, return_counts=True)
    if torch.all(counts==1) and torch.numel(intratrace)==0: ## no duplicates in label
        #### this needs to take into account intratraces!!!
        return index, data, shape_ 
    else: ## slice and wellorder after then

        ## row-wise slice
        for uni in uniques[counts>1]:
            lookat = torch.where(uni==label)[0]

            expand = index[lookat[0]].unsqueeze(0).expand_as(index[lookat[1:]])
            result = (index[lookat[1:]] == expand)
            I      = torch.all(result, dim=0)
            index  = index[:,I] ## row-wise-chop
            data   = data[I]    ## row-wise-chop
        
        ## column-wise slice, intra-external indices
        keep  = first_occurrence_mask(label)
        index = index[keep] ## chop list-of-tuples
        label = label[keep] ## chop label
        shape_= shape_[keep]

        ## column-wise slice, remove intratrace
        if len(intratrace)!=0: 
            for j in intratrace: ## each column
                keep   = (label!=j)
                label  = label[keep]
                index  = index[keep]
                shape_ = shape_[keep]
        else:
            None
    

        if torch.numel(index)==0: #### IF COMPLETELY REMOVED, trivial sparse-tensor "sparse-scalar"
            return torch.zeros((1,1), device=index.device, dtype=index.dtype), torch.sum(torch.unsqueeze(data, 0), dim=1), torch.ones(1, device=shape_.device, dtype=shape_.dtype) #
            #return torch.zeros((1,1),device=index.device), torch.sum(torch.unsqueeze(data, 0), dim=1), torch.ones(1, device=data.device)
            #return torch.zeros_like( torch.unsqueeze(index[:,0], 0)[:,0] , dtype=index.dtype),  data, torch.ones_like( torch.unsqueeze(index[:,0], 0)[:,0] , dtype=shape_.dtype) 

        else: ### sum over domains
            ################ 
            ### lex-order index arrayXX
            i     = lexsort(index)
            index = index[:,i]
            data  = data[i]

            domains = torch.concat([torch.ones_like(label[0].reshape((1)), dtype=torch.int32), torch.any((torch.diff( index , dim=1) != 0), dim=0).long()])

            ## do indices
            index = index[:,domains!=0]

            ## do data
            domains = (torch.cumsum(domains, 0)-1)
            ZZ      = torch.zeros_like(index[0], dtype=data.dtype)
            data    = ZZ.scatter_add(0, domains, data) 

        return index, data, shape_
    
@torch.jit.script
def spa_post_intraTr_(index, data, shape_, label):

    uniques, counts = torch.unique(label, return_counts=True)
    if torch.all(counts==1): ## no duplicates in label
        return index, data, shape_
    else: ## slice and wellorder after then

        ## row-wise slice
        for uni in uniques[counts>1]:
            lookat = torch.where(uni==label)[0]

            expand = index[lookat[0]].unsqueeze(0).expand_as(index[lookat[1:]])
            result = (index[lookat[1:]] == expand)
            I      = torch.all(result, dim=0)
            index  = index[:,I] ## row-wise-chop
            data   = data[I]    ## row-wise-chop
        
        ## column-wise slice, intra-external indices
        keep  = first_occurrence_mask(label)
        index = index[keep] ## chop list-of-tuples
        label = label[keep] ## chop label
        shape_= shape_[keep]
        
        if torch.numel(index)!=0: ## index is empty
            #print(index.shape)
            i     = lexsort(index)
            index = index[:,i]
            data  = data[i]

            domains = torch.concat([torch.ones_like(label[0].reshape((1)), dtype=torch.int32), torch.any((torch.diff( index , dim=1) != 0), dim=0).long()])

            ## do indices
            index = index[:,domains!=0]

            ## do data
            domains = (torch.cumsum(domains, 0)-1)
            ZZ      = torch.zeros_like(index[0], dtype=data.dtype)
            data    = ZZ.scatter_add(0, domains, data)

            return index, data, shape_
        else:
            #data = torch.unsqueeze(torch.unsqueeze(data, 0),0) ## !!!!
            #return torch.zeros_like(data[:,:,0], dtype=torch.int64), torch.sum(data), torch.ones_like(data[:,0,0]) ###!!!
            return torch.zeros((1,1), device=index.device, dtype=index.dtype), torch.sum(torch.unsqueeze(data, 0), dim=1), torch.ones(1, device=shape_.device, dtype=shape_.dtype) #


#@torch.jit.script
def spa_post_trans_(index, data, shape_, rhs, RHS):
    rhs=rhs[first_occurrence_mask(rhs)]
    j = torch.argsort(rhs)
    k = torch.argsort(RHS)
    ik= iargsort(k)
    m = j[ik]

    cc= index[m,:]
    s = shape_[m]
    n = lexsort( cc )

    return torch.sparse_coo_tensor( cc[:,n] , data[n], [int(i.item()) for i in s]) ## .type(torch.LongTensor) not needed



### TESTS
#from uniform_random_sparse_tensor import uniform_random_sparse_tensor

#shape = torch.tensor([2,2,2,3])
#A = uniform_random_sparse_tensor(torch.prod(shape), shape)
#A = torch.sparse_coo_tensor( torch.concat((A._indices(),A._indices()),axis=1) , torch.concat((A._values(),A._values())), A.shape)
#print(sp_tensor_intraTr(A, torch.tensor([2, 2, 3, 7]), intratrace=torch.tensor([2])) )


"""
tensordot.py

This generalizes tensordot for sparse tensors (with dense tensors).
"""

import torch
from .generic_functions import ints_to_tuples, cartesian_product
from .tensor_to_matrix import sp_tensor_to_matrix, dn_tensor_to_matrix, rawsp_tensor_to_matrix

@torch.jit.script
def sd_outer(a, b):
    """
    GIVEN : a (sparse)
            b (dense)
    GET   : c   (torch.tensor{dense})
    """
    if a.is_sparse:
        a=a.to_dense()
    else:
        if b.is_sparse:
            b=b.to_dense()
        else:
            torch.tensordot(a, b, dims=(0))
    return  torch.tensordot(a, b, dims=(0))

@torch.jit.script
def sd_outer_(a_index, a_data, a_shape, b):
    """
    GIVEN : a (sparse)
            b (dense)
    GET   : c   (torch.tensor{dense})
    """
    a = torch.sparse_coo_tensor(a_index, a_data, [int(q.item()) for q in a_shape])
    a=a.to_dense()
    return  torch.tensordot(a, b, dims=(0))

@torch.jit.script
def ds_outer_(a, b_index, b_data, b_shape):
    """
    GIVEN : a (sparse)
            b (dense)
    GET   : c   (torch.tensor{dense})
    """
    b = torch.sparse_coo_tensor(b_index, b_data, [int(q.item()) for q in b_shape])
    b=b.to_dense()
    return torch.tensordot(a, b, dims=(0))

@torch.jit.script
def sparse_outer(a,b):
    """
    GIVEN : a,b (torch.sparse_coo_tensor)
    GET   : c   (torch.sparse_coo_tensor)
    """
    data = torch.outer(a._values(), b._values()).reshape(-1)
    inde = cartesian_product(a._indices(),b._indices())
    size = torch.concat((torch.tensor(a.shape, device=a.device),torch.tensor(b.shape, device=b.device)))
    size = [int(elem.item()) for elem in size]
    return torch.sparse_coo_tensor(inde, data, size) #return [inde, data, size]

@torch.jit.script
def sd_tensordot(a, b, dims=None):
    """
    GIVEN : a, sparse torch.tensor
            b, dense  torch.tensor
            dims (2d-int-torch.tensor, with 0th axis being 2: e.g. shape=(2,4))
    GET   : c (torch.tensor{dense} XOR torch.tensor{sparse} is just direct-product)
    """
    if dims.numel()==0: ## directproduct
        c = sd_outer(a, b)
    else: ## directintersection
        if (not a.is_sparse) and (b.is_sparse): # a is dense && b is sparse
            m_A, exlabel_A = dn_tensor_to_matrix(a, dims[0], left=False ) ## True
            m_B, exlabel_B = sp_tensor_to_matrix(b, dims[1], left=True ) ## False
            c = m_B @ m_A ### SPARSE is first??!?
            c = c.T.reshape( [ int(i) for i in torch.concat([torch.flatten(exlabel_B), torch.flatten(exlabel_A)])])
        else: ## b is dense
            m_A, exlabel_A = sp_tensor_to_matrix(a, dims[0], left=True ) ## True
            m_B, exlabel_B = dn_tensor_to_matrix(b, dims[1], left=False  ) ## False
            c = m_A @ m_B ## Sparse is 1st
            c = c.reshape([int(i) for i in torch.concat([exlabel_A, exlabel_B])])
    return c

@torch.jit.script
def ds_tensordot(a, b, dims=None):
    """
    GIVEN : a, dense  torch.tensor
            b, sparse torch.tensor
            dims (2d-int-torch.tensor, with 0th axis being 2: e.g. shape=(2,4))
    GET   : c (torch.tensor{dense} XOR torch.tensor{sparse} is just direct-product)
    """
    if dims.numel()==0: ## directproduct
        c = sd_outer(a, b)
    else: ## directintersection
        if (not a.is_sparse) and (b.is_sparse): # a is dense && b is sparse
            m_A, exlabel_A = dn_tensor_to_matrix(a, dims[0], left=False ) ## True
            m_B, exlabel_B = sp_tensor_to_matrix(b, dims[1], left=True ) ## False
            c = m_B @ m_A ### SPARSE is first??!?
            c = c.T.reshape( [ int(i) for i in torch.concat([torch.flatten(exlabel_B), torch.flatten(exlabel_A)])])
        else: ## b is dense
            m_A, exlabel_A = sp_tensor_to_matrix(a, dims[0], left=True ) ## True
            m_B, exlabel_B = dn_tensor_to_matrix(b, dims[1], left=False  ) ## False
            c = m_A @ m_B ## Sparse is 1st
            c = c.reshape([int(i) for i in torch.concat([exlabel_A, exlabel_B])])
    return c

@torch.jit.script
def ss_tensordot(a, b, dims=None):
    """
    GIVEN : a,b (torch.tensor{sparse})
            dims (2d-int-torch.tensor, with 0th axis being 2: e.g. shape=(2,4))
    GET   : c (torch.tensor{sparse})
    """
    if dims.numel()==0: ## directproduct 
        c = sparse_outer(a,b)
    else: ## directintersection
        m_A, exlabel_A = sp_tensor_to_matrix(a, dims[0], left=True ) ## True
        m_B, exlabel_B = sp_tensor_to_matrix(b, dims[1], left=False ) ## False

        if torch.numel(exlabel_A)==0 and torch.numel(exlabel_B)!=0:
            c = m_A @ m_B ## let it be a....
            I = ints_to_tuples(c._indices()[1], exlabel_B)
            c = torch.sparse_coo_tensor(I, c._values(), [int(i) for i in exlabel_B])
        else:
            if torch.numel(exlabel_B)==0 and torch.numel(exlabel_A)!=0:
                c = m_A @ m_B   ## B side is null, result is a "vector" in A
                I = ints_to_tuples(c._indices()[0], exlabel_A)
                c = torch.sparse_coo_tensor(I, c._values(), [int(i) for i in exlabel_A])
            else:
                if torch.numel(exlabel_B)!=0 and torch.numel(exlabel_A)!=0:
                    c = m_A @ m_B ## A side is null, result is a "vector" in B
                    I = torch.concat([ints_to_tuples(c._indices()[0], exlabel_A), ints_to_tuples(c._indices()[1], exlabel_B)])
                    c = torch.sparse_coo_tensor(I, c._values(), [int(i) for i in torch.concat([exlabel_A, exlabel_B])])
                else: ## A & B side is null, result is scalar, a trivial sparse tensor
                    if torch.numel(exlabel_B)==0 and torch.numel(exlabel_A)==0:
                        c = m_A @ m_B ## let it be a....
                    else:
                        raise ValueError
    return c

@torch.jit.script
def ss_tensordot_(a_index, a_data, a_shape, b_index, b_data, b_shape, dims=None):
    """
    GIVEN : a,b (torch.tensor{sparse})
            dims (2d-int-torch.tensor, with 0th axis being 2: e.g. shape=(2,4))
    GET   : c (torch.tensor{sparse})
    """
    if dims.numel()==0: ## directproduct
        I     = cartesian_product( a_index, b_index )
        data  = torch.outer( a_data, b_data ).reshape(-1)
        shape_= torch.concat([a_shape, b_shape])
    else: ## directintersection
        m_A, exlabel_A = rawsp_tensor_to_matrix(a_index, a_data, a_shape, dims[0], left=True )
        m_B, exlabel_B = rawsp_tensor_to_matrix(b_index, b_data, b_shape, dims[1], left=False)

        if torch.numel(exlabel_A)==0 and torch.numel(exlabel_B)!=0:
            c = m_A @ m_B ## let it be a....
            I = ints_to_tuples(c._indices()[1], exlabel_B)
            data   = c._values()
            shape_ = exlabel_B
        else:
            if torch.numel(exlabel_B)==0 and torch.numel(exlabel_A)!=0:
                c = m_A @ m_B   ## B side is null, result is a "vector" in A
                I = ints_to_tuples(c._indices()[0], exlabel_A)
                data   = c._values()
                shape_ = exlabel_A
            else:
                if torch.numel(exlabel_B)!=0 and torch.numel(exlabel_A)!=0:
                    c = m_A @ m_B ## A side is null, result is a "vector" in B
                    I = torch.concat([ints_to_tuples(c._indices()[0], exlabel_A), ints_to_tuples(c._indices()[1], exlabel_B)])
                    data   = c._values()
                    shape_ = torch.concat([exlabel_A, exlabel_B])
                else: ## A & B side is null, result is scalar, a trivial sparse tensor
                    if torch.numel(exlabel_B)==0 and torch.numel(exlabel_A)==0:
                        c = m_A @ m_B ## let it be a....
                        I      = torch.zeros_like( torch.unsqueeze(a_index[:,0], 0)[:,0] , dtype=a_index.dtype)
                        data   = c._values()
                        shape_ = torch.ones_like( torch.unsqueeze(a_index[:,0], 0)[:,0] , dtype=a_shape.dtype) 
                    else:
                        raise ValueError
    return I, data, shape_

@torch.jit.script
def ds_tensordot_(a, b_index, b_data, b_shape, dims=None):
    """
    GIVEN : a, dense  torch.tensor
            b, sparse torch.tensor
            dims (2d-int-torch.tensor, with 0th axis being 2: e.g. shape=(2,4))
    GET   : c (torch.tensor{dense} XOR torch.tensor{sparse} is just direct-product)
    """
    if dims.numel()==0: ## directproduct
        c = ds_outer_(a, b_index, b_data, b_shape)
    else: ## directintersection
        # a is dense && b is sparse
        m_A, exlabel_A = dn_tensor_to_matrix(a, dims[0], left=False ) ## True
        m_B, exlabel_B = rawsp_tensor_to_matrix(b_index, b_data, b_shape, dims[1], left=False)
        
        #m_B, exlabel_B = sp_tensor_to_matrix(b, dims[1], left=True ) ## False
        c = m_B @ m_A ### SPARSE is first??!?
        c = c.T.reshape( [ int(i) for i in torch.concat([torch.flatten(exlabel_B), torch.flatten(exlabel_A)])])
    return c

@torch.jit.script
def sd_tensordot_(a_index, a_data, a_shape, b, dims=None):
    """
    GIVEN : a, sparse torch.tensor
            b, dense  torch.tensor
            dims (2d-int-torch.tensor, with 0th axis being 2: e.g. shape=(2,4))
    GET   : c (torch.tensor{dense} XOR torch.tensor{sparse} is just direct-product)
    """
    if dims.numel()==0: ## directproduct
        c = sd_outer_(a_index, a_data, a_shape, b)
    else: ## directintersection
        m_A, exlabel_A = rawsp_tensor_to_matrix(a_index, a_data, a_shape, dims[0], left=True ) ## True
        m_B, exlabel_B = dn_tensor_to_matrix(b, dims[1], left=False  ) ## False
        c = m_A @ m_B ## Sparse is 1st
        c = c.reshape([int(i) for i in torch.concat([exlabel_A, exlabel_B])])
    return c


from distutils.core import setup

with open("README.md") as readme:
    long_description = readme.read()

setup(
  name = 'bisum',         # How you named your package folder (MyLib)
  packages = ['bisum'],   # Chose the same as "name"
  version = '0.1.1',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = "binary sparse and dense tensor partial-tracing",   # Give a short description about your library
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = 'Julio Candanedo',                   # Type in your name
  author_email = 'juliojcandanedo@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/jcandane/bisum/',   # Provide either the link to your github or to your website
  keywords = ['pytorch', 'torch', 'tensors', 'Sparse Tensor', 'Sparse', 'contraction', 'partial-tracing', 'einsum', 'tensordot','attention'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'torch'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',
  ],
)


"""
tensor_to_matrix.py

Fold dense and sparse tensors into matrices (dense or sparse)
"""

import torch
from .generic_functions import pytorch_delete, tuples_to_ints


@torch.jit.script
def rawsp_tensor_to_matrix(index, data, shape_, adj_matrix, left : bool = True):
    """ !!!! tensor_in should instead be:    indices, data, shape (as tensors), because by DEAFULT SparseTensors change index types to int64 :(
    *** Fold a Tensor with given external xor internal indices into a external-internal matrix
    GIVEN : index (2d-int-torch.tensor)
            data (1d-torch.tensor)
            shape_ (1d-int-torch.tensor)
            adj_matrix (2d-int-torch.tensor)
            *left (side in-which external-indices will be placed on)
    GET   : matrix (torch.tensor[2d-dense])
            externals-shape (1d-int-torch.tensor, prefolding)
    """

    shaper = [ shape_[i] + 0*adj_matrix[0].reshape((1,1)) for i in range(len(shape_)) ]
    arange = [ adj_matrix[0].reshape((1,1)) for i in range(len(shape_)) ]
    arange = ( torch.cumsum( (1+0*torch.concat(arange))[:,0] , 0 ) - 1 ) ## arange

    internal_index = adj_matrix
    external_index = pytorch_delete( arange , internal_index) ### !!!arange not in Everything not included....
    shape = torch.flatten(torch.concat(shaper)) #!!! #torch.tensor(tensor_in.shape) ### !!!arange not in 
    
    ### LEFT side vector!!  need RIGHT side too!!!!
    if adj_matrix.shape[0]==index.shape[0]: ### then reshape to a vector, no external-indices
        if left:
            sA = [ 1, torch.prod(shape[internal_index]).item() ]
            sA = [int( s ) for s in sA]
            I  = tuples_to_ints( index[internal_index,:], shape[internal_index] )
            I0 = torch.zeros_like( I , dtype=I.dtype)
            I  = torch.stack((I0, I))
            matrix = torch.sparse_coo_tensor( I , data, sA )
        else:
            sA = [ torch.prod(shape[internal_index]).item(), 1] #[1, ... , 1]
            sA = [int( s ) for s in sA]
            I  = tuples_to_ints( index[internal_index,:], shape[internal_index] )
            I0 = torch.zeros_like( I , dtype=I.dtype)
            I  = torch.stack((I, I0))
            matrix = torch.sparse_coo_tensor( I , data, sA )

    else: ### this creates an internal/external-matrix
        if left:
            sA = torch.concat([torch.prod(shape[external_index].reshape((1,external_index.shape[0])), dim=1), torch.prod(shape[internal_index].reshape((1,internal_index.shape[0])), dim=1)])
            sA = [int( s.item() ) for s in sA]
            I    = tuples_to_ints( index[internal_index,:], shape[internal_index] ) ### check devices for shape & tensor_in
            E    = tuples_to_ints( index[external_index,:], shape[external_index] )
            EI   = torch.stack([E, I], dim=0)
            matrix = torch.sparse_coo_tensor( EI , data, sA )

        else: ## external-indices on right-side
            sA = torch.concat([torch.prod(shape[internal_index].reshape((1,internal_index.shape[0])), dim=1), torch.prod(shape[external_index].reshape((1,external_index.shape[0])), dim=1)])
            sA = [int( s.item() ) for s in sA]
            I    = tuples_to_ints( index[internal_index,:], shape[internal_index] )
            E    = tuples_to_ints( index[external_index,:], shape[external_index] )
            IE   = torch.stack([I, E], dim=0)
            matrix = torch.sparse_coo_tensor( IE , data, sA )

    return matrix, shape[external_index]


@torch.jit.script
def sp_tensor_to_matrix(tensor_in, adj_matrix, left : bool = True):
    """
    *** Fold a Tensor with given external xor internal indices into a external-internal matrix
    GIVEN : tensor_in (torch.tensor{sparse})
            adj_matrix (2d-int-torch.tensor)
            *left (side in-which external-indices will be placed on)
    GET   : matrix (torch.tensor[2d-dense])
            externals-shape (1d-int-torch.tensor, prefolding)
    """

    shaper = [ tensor_in.shape[i] + 0*adj_matrix[0].reshape((1,1)) for i in range(len(tensor_in.shape)) ]
    arange = [ adj_matrix[0].reshape((1,1)) for i in range(len(tensor_in.shape)) ]
    arange = ( torch.cumsum( (1+0*torch.concat(arange))[:,0] , 0 ) - 1 ) ## arange

    internal_index = adj_matrix
    external_index = pytorch_delete( arange , internal_index) ### !!!arange not in Everything not included....
    shape = torch.flatten(torch.concat(shaper)) #!!! #torch.tensor(tensor_in.shape) ### !!!arange not in 
    
    ### LEFT side vector!!  need RIGHT side too!!!!
    if adj_matrix.shape[0]==tensor_in.ndim: ### then reshape to a vector, no external-indices
        if left:
            sA = [ 1, torch.prod(shape[internal_index]).item() ]
            sA = [int( s ) for s in sA]
            I  = tuples_to_ints( tensor_in._indices()[internal_index,:], shape[internal_index] )
            I0 = torch.zeros_like( I , dtype=I.dtype)
            I  = torch.stack((I0, I))
            matrix = torch.sparse_coo_tensor( I , tensor_in._values(), sA )
        else:
            sA = [ torch.prod(shape[internal_index]).item(), 1] #[1, ... , 1]
            sA = [int( s ) for s in sA]
            I  = tuples_to_ints( tensor_in._indices()[internal_index,:], shape[internal_index] )
            I0 = torch.zeros_like( I , dtype=I.dtype)
            I  = torch.stack((I, I0))
            matrix = torch.sparse_coo_tensor( I , tensor_in._values(), sA )

    else: ### this creates an internal/external-matrix
        if left:
            sA = torch.concat([torch.prod(shape[external_index].reshape((1,external_index.shape[0])), dim=1), torch.prod(shape[internal_index].reshape((1,internal_index.shape[0])), dim=1)])
            sA = [int( s.item() ) for s in sA]
            I    = tuples_to_ints( tensor_in._indices()[internal_index,:], shape[internal_index] ) ### check devices for shape & tensor_in
            E    = tuples_to_ints( tensor_in._indices()[external_index,:], shape[external_index] )
            EI   = torch.stack([E, I], dim=0)
            matrix = torch.sparse_coo_tensor( EI , tensor_in._values(), sA )

        else: ## external-indices on right-side
            sA = torch.concat([torch.prod(shape[internal_index].reshape((1,internal_index.shape[0])), dim=1), torch.prod(shape[external_index].reshape((1,external_index.shape[0])), dim=1)])
            sA = [int( s.item() ) for s in sA]
            I    = tuples_to_ints( tensor_in._indices()[internal_index,:], shape[internal_index] )
            E    = tuples_to_ints( tensor_in._indices()[external_index,:], shape[external_index] )
            IE   = torch.stack([I, E], dim=0)
            matrix = torch.sparse_coo_tensor( IE , tensor_in._values(), sA )

    return matrix, shape[external_index]

@torch.jit.script
def dn_tensor_to_matrix(tensor_in, adj_matrix, left : bool = True):
    """
    *** Fold a Tensor with given external xor internal indices into a external-internal matrix
    GIVEN : tensor_in (torch.tensor)
            adj_matrix (2d-int-torch.tensor)
            *left (side in-which external-indices will be placed on)
    GET   : matrix (torch.tensor[2d-dense])
            externals-shape (1d-int-torch.tensor, prefolding)
    """

    shaper = [ tensor_in.shape[i] + 0*adj_matrix[0].reshape((1,1)) for i in range(len(tensor_in.shape)) ]
    arange = [ adj_matrix[0].reshape((1,1)) for i in range(len(tensor_in.shape)) ]
    arange =   torch.cumsum( (1+0*torch.concat(arange))[:,0] , 0 ) - 1 ## arange

    internal_index = adj_matrix
    external_index = pytorch_delete( arange , internal_index) ### !!!arange not in Everything not included....
    shape = torch.concat(shaper) #torch.tensor(tensor_in.shape) ### !!!arange not in 
    
    if adj_matrix.shape[0]==tensor_in.ndim: ### then reshape to a vector, no external-indices
        if left:
            sA = [ torch.prod(shape[internal_index]).item(), 1]
            sA = [int( s ) for s in sA]
            permute = [int(elem.item()) for elem in internal_index]
            matrix  = torch.permute(tensor_in, permute).reshape(-1)
        else:
            sA = [1, torch.prod(shape[internal_index]).item()]
            sA = [int( s ) for s in sA]
            permute = [int(elem.item()) for elem in internal_index]
            matrix  = torch.permute(tensor_in, permute).reshape(-1)

    else: ### this creates an internal/external-matrix
        if left:
            sA = torch.concat([torch.prod(shape[external_index].reshape((1,external_index.shape[0])), dim=1), torch.prod(shape[internal_index].reshape((1,internal_index.shape[0])), dim=1)])
            sA = [int( s.item() ) for s in sA]
            permute = [int(elem.item()) for elem in torch.concat( [external_index, internal_index] )]
            matrix  = torch.permute(tensor_in, permute).reshape(sA)

        else: ## external-indices on right-side
            sA = torch.concat([torch.prod(shape[internal_index].reshape((1,internal_index.shape[0])), dim=1), torch.prod(shape[external_index].reshape((1,external_index.shape[0])), dim=1)])
            sA = [int( s.item() ) for s in sA]
            permute = [int(elem.item()) for elem in torch.concat( [internal_index, external_index] )]
            matrix  = torch.permute(tensor_in, permute).reshape( sA )

    return matrix, shape[external_index]


"""
generic_functions.py

This python file contains defs/functions that could be widely used in many codes.
"""

import torch

###################################
######## INVERSE ARGSORT  #########
###################################

@torch.jit.script
def iargsort(i):
    """
    ***suppose we have a unsorted-array A, its args to sort are obtained via:
    i = argsort(A)
    *its this indices 'i' that are used below, to obtain the inverse, i.e.
    A_sorted = A[i]
    A_back   = A_sorted[i_rev] = A[i][i_rev]
    ***
    GIVEN : i (argsort-indices)
    GET   : i_rev ()
    """
    i_rev    = torch.ones_like(i, dtype=i.dtype)
    i_rev[i] = torch.arange(len(i), dtype=i.dtype, device=i.device)
    return i_rev

###################################
######### 1st Occurrence ##########
###################################

@torch.jit.script
def first_occurrence_mask(tensor):
    j = torch.argsort(tensor)
    i = iargsort(j)
    return (torch.concat((torch.ones(1, dtype=tensor.dtype, device=tensor.device),torch.diff( tensor[j] ))) != 0)[i]

###################################
######### PyTorch DELETE ##########
###################################

@torch.jit.script
def pytorch_delete(data, args_to_delete):
    """
    *** delete entries of a torch.tensor given indices (to_delete)
    GIVEN : torch.tensor
    GET   : torch.tensor (without deleted entries)
    """
    mask = torch.ones_like(data, dtype=torch.bool)
    mask[args_to_delete] = False
    return data[mask]

###################################
############# LEXSORT #############
###################################

@torch.jit.script
def lexsort(LoT):
    """
    lexicographically-order list-of-tuples (tuple-index, list-index), e.g. (6,123854)
    such that the 0-th entry in LoT is the most important, i.e. (a_0,a_1,a_2,...,a_n)
    GIVEN : LoT (2d torch.tensor)
    GET   : idx (1d torch.tensor : indices to lexsort LoT)
    """
    idx = torch.argsort(LoT[-1], stable=True)
    for k in reversed(LoT[:-1]):
        idx = idx.gather(0, torch.argsort(k.gather(0, idx), stable=True))
    return idx

@torch.jit.script
def nplexsort(LoT):
    """ same as numpy.lexsort a.k.a colexsort
    lexicographically-order list-of-tuples (tuple-index, list-index), e.g. (6,123854)
    such that the n-th entry in LoT is the most important, i.e. (a_0,a_1,a_2,...,a_n)
    GIVEN : LoT (2d torch.tensor)
    GET   : idx (1d torch.tensor : indices to lexsort LoT)
    """
    idx = torch.argsort(LoT[0], stable=True)
    for k in LoT[1:]: ## each axis after 0th
        idx = idx.gather(0, torch.argsort(k.gather(0, idx), stable=True))
    return idx

###################################
##### LEXICOGRAPHIC ORDERING ######
###################################

@torch.jit.script
def lex_lessthan(a, b): ## a<b ? for tuples a & b
    c = b - a
    C = torch.nonzero(c)
    if C.numel()==0: ## special-case: if-equal?
        return False
    else:
        if (c[C[0]] > 0): ## is 1st nonzero positive?
            return True
        else:
            return False

###################################
######### SPARSE RESHAPER #########
###################################

@torch.jit.script
def ints_to_tuples(the_ints, denseshape):
    """
    ***Reshape single sparse list-of-tuples 1d tensor/array into a 2d array/tensor list-of-tuples.
    GIVEN : the_ints (int torch.tensor)
            denseshape (int 1d torch.tensor)
    GET   : 2d torch.tensor (being the 'list'-of-tuples)
    """
    the_ints   = torch.flatten(the_ints) ## forces the_ints to be 1-dimensional
    denseshape = torch.flatten(denseshape)
    denseshape = torch.concat( (torch.flip( torch.cumprod( torch.roll(torch.flip(denseshape, [0]), 1)[1:] , dim=0 ), [0] ), torch.ones_like(denseshape.reshape((denseshape.shape[0],1))[0], dtype=torch.int)),0)
    out_tuple  = []

    for s in denseshape: ## for each column in new-shape, generates tuple....
        out_tuple.append( the_ints // s )
        the_ints = torch.remainder( the_ints, s )
    return torch.stack(out_tuple, 0)

@torch.jit.script
def tuples_to_ints(list_of_tuples, denseshape):
    """
    ***Reshape sparse list-of-tuples tensor/array into a 1d array/tensor.
    GIVEN:  list_of_tuples (2d int torch.tensor, with shape (col'n,rows), eg (3,12357))
            dense_shape (1d torch.tensor, shape of the dense representation)
    GET:    1d int torch.tensor (corresponding to pair-function)
    """
    denseshape = torch.flatten(denseshape)
    if list_of_tuples.shape[0]!=denseshape.shape[0]:
        if torch.numel(list_of_tuples)==0:
            raise ValueError("list_of_tuples is empty")
        else:
            raise ValueError("tuple-shapes and shape must match")

    denseshape = torch.concat( (torch.flip( torch.cumprod( torch.roll(torch.flip(denseshape, [0]), 1)[1:] , dim=0 ), [0] ), torch.ones_like(denseshape.reshape((denseshape.shape[0],1))[0], dtype=denseshape.dtype)),0)
    denseshape = denseshape.type(list_of_tuples.dtype)
    #return torch.matmul(denseshape, list_of_tuples ) ## vector @ matrix product, (n) @ (n, N) , with O ~ nN (linear in N) ### this isnt implemented in cuda for pytorch!!!!! only works for int32
    return torch.sum(denseshape.unsqueeze(1) * list_of_tuples, 0) ## matrix-product
    #return torch.matmul(list_of_tuples.T, denseshape ) 

#### TESTS 
#device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
#device = torch.device("cpu")
#print(device)

#A       = torch.reshape(torch.randint(0, 15, (18,), device=device), (1,1,18))
#shaper  = torch.tensor([2,4,6], device=device)
#print( torch.allclose( torch.flatten(A), tuples_to_ints(ints_to_tuples(A, shaper), shaper)) )

##### tester!!!
#device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
#print(device)
#B = torch.reshape(torch.randint(0, 15, (18,), device=device, dtype=torch.int32), (1,1,18))
#sh= torch.tensor([2,4,6], device=device, dtype=torch.int32)

#print( torch.allclose( torch.flatten(A), tuples_to_ints(ints_to_tuples(A, shaper), shaper)) )

##########################################
############## OUTERPRODUCT ##############
##########################################

@torch.jit.script
def cartesian_product(a, b):
    ab = (a.unsqueeze(1) * torch.ones(b.shape[1], dtype=a.dtype, device=a.device).unsqueeze(1) ).swapaxes(1,2).reshape((a.shape[0], a.shape[1] * b.shape[1]))
    ba = (b.unsqueeze(1) * torch.ones(a.shape[1], dtype=b.dtype, device=a.device).unsqueeze(1) ).reshape((b.shape[0], b.shape[1] * a.shape[1]))
    return torch.cat((ab, ba), dim=0)
